/*
 *  Copyright 2010 Chris Pheby
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.jadira.bindings.core;

import java.io.File;
import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.InetAddress;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.Currency;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.TimeZone;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

import javax.enterprise.inject.Typed;

import org.jadira.bindings.core.annotation.BindingScope;
import org.jadira.bindings.core.annotation.DefaultBinding;
import org.jadira.bindings.core.api.FromStringUnmarshaller;
import org.jadira.bindings.core.api.StringBinding;
import org.jadira.bindings.core.api.ToStringMarshaller;
import org.jadira.bindings.core.general.CompositeStringBinding;
import org.jadira.bindings.core.general.ConstructorFromStringUnmarshaller;
import org.jadira.bindings.core.general.FromStringBinding;
import org.jadira.bindings.core.general.MethodFromStringUnmarshaller;
import org.jadira.bindings.core.general.MethodToStringMarshaller;
import org.jadira.bindings.core.general.ToStringBinding;
import org.jadira.bindings.core.jdk.AtomicBooleanStringBinding;
import org.jadira.bindings.core.jdk.AtomicIntegerStringBinding;
import org.jadira.bindings.core.jdk.AtomicLongStringBinding;
import org.jadira.bindings.core.jdk.BigDecimalStringBinding;
import org.jadira.bindings.core.jdk.BigIntegerStringBinding;
import org.jadira.bindings.core.jdk.BooleanStringBinding;
import org.jadira.bindings.core.jdk.ByteStringBinding;
import org.jadira.bindings.core.jdk.CalendarStringBinding;
import org.jadira.bindings.core.jdk.CharSequenceStringBinding;
import org.jadira.bindings.core.jdk.CharacterStringBinding;
import org.jadira.bindings.core.jdk.ClassStringBinding;
import org.jadira.bindings.core.jdk.CurrencyStringBinding;
import org.jadira.bindings.core.jdk.DateStringBinding;
import org.jadira.bindings.core.jdk.DoubleStringBinding;
import org.jadira.bindings.core.jdk.FileStringBinding;
import org.jadira.bindings.core.jdk.FloatStringBinding;
import org.jadira.bindings.core.jdk.InetAddressStringBinding;
import org.jadira.bindings.core.jdk.IntegerStringBinding;
import org.jadira.bindings.core.jdk.LocaleStringBinding;
import org.jadira.bindings.core.jdk.LongStringBinding;
import org.jadira.bindings.core.jdk.PackageStringBinding;
import org.jadira.bindings.core.jdk.ShortStringBinding;
import org.jadira.bindings.core.jdk.StringBufferStringBinding;
import org.jadira.bindings.core.jdk.StringBuilderStringBinding;
import org.jadira.bindings.core.jdk.StringStringBinding;
import org.jadira.bindings.core.jdk.TimeZoneStringBinding;
import org.jadira.bindings.core.jdk.URIStringBinding;
import org.jadira.bindings.core.jdk.URLStringBinding;
import org.jadira.bindings.core.jdk.UUIDStringBinding;
import org.jadira.bindings.core.loader.Binding;
import org.jadira.bindings.core.loader.BindingConfiguration;
import org.jadira.bindings.core.loader.BindingXmlLoader;
import org.jadira.bindings.core.loader.Provider;
import org.jadira.bindings.core.spi.BindingProvider;
import org.jadira.bindings.core.utils.lang.IterableEnumeration;
import org.jadira.bindings.core.utils.reflection.ClassLoaderUtils;

/**
 * Core binding capability.
 * Typically you would obtain an instance of this class either
 * by direct construction, or in JDK6 and above by using 
 * <code>java.util.ServiceLoader</code>.
 * 
 * Default construction of this class registers bindings for core
 * JDK classes, as well as any bindings specified using bindings.xml
 * 
 * You can also register additional bindings programmatically.
 * @see StringBinding StringBinding is used to define a qualifier
 */
@Typed({})
public class BasicStringBinder implements StringBinder {

    private final Set<BindingProvider> bindingProviders = new HashSet<BindingProvider>();
    
    private final ConcurrentHashMap<Class<?>, ConcurrentHashMap<Class<? extends Annotation>, ToStringMarshaller<?>>> registeredMarshalMethods = new ConcurrentHashMap<Class<?>, ConcurrentHashMap<Class<? extends Annotation>, ToStringMarshaller<?>>>();
    
    private final ConcurrentHashMap<Class<?>, ConcurrentHashMap<Class<? extends Annotation>, FromStringUnmarshaller<?>>> registeredUnmarshalMethods = new ConcurrentHashMap<Class<?>, ConcurrentHashMap<Class<? extends Annotation>, FromStringUnmarshaller<?>>>();
    
    private final ConcurrentHashMap<Class<?>, Set<Class<? extends Annotation>>> annotationsApplied = new ConcurrentHashMap<Class<?>, Set<Class<? extends Annotation>>>();
    
    public BasicStringBinder() {
        this(true);
    }

    public BasicStringBinder(boolean includeBuiltInBindings) {

        if (includeBuiltInBindings) {

            initJdkBindings();

            // Before we load other configuration, load the binding
            // configuration from
            // jadira-bindings' jar to ensure no-one overrides our built-in
            // mappings.
            initBuiltInBindings();

            // Init bindings registered from other jars
            initRegisteredBindings();
        }
    }

    private void initJdkBindings() {
        
        register(AtomicBoolean.class, new AtomicBooleanStringBinding());
        register(AtomicInteger.class, new AtomicIntegerStringBinding());
        register(AtomicLong.class, new AtomicLongStringBinding());
        register(BigDecimal.class, new BigDecimalStringBinding());
        register(BigInteger.class, new BigIntegerStringBinding());
        register(Boolean.class, new BooleanStringBinding());
        register(Byte.class, new ByteStringBinding());
        register(Calendar.class, new CalendarStringBinding());
        register(Character.class, new CharacterStringBinding());
        register(CharSequence.class, new CharSequenceStringBinding());
        register(Class.class, new ClassStringBinding());
        register(Currency.class, new CurrencyStringBinding());
        register(Date.class, new DateStringBinding());
        register(Double.class, new DoubleStringBinding());
        register(File.class, new FileStringBinding());
        register(Float.class, new FloatStringBinding());
        register(InetAddress.class, new InetAddressStringBinding());
        register(Integer.class, new IntegerStringBinding());
        register(Locale.class, new LocaleStringBinding());
        register(Long.class, new LongStringBinding());
        register(Package.class, new PackageStringBinding());
        register(Short.class, new ShortStringBinding());
        register(StringBuffer.class, new StringBufferStringBinding());
        register(StringBuilder.class, new StringBuilderStringBinding());
        register(String.class, new StringStringBinding());
        register(TimeZone.class, new TimeZoneStringBinding());
        register(URI.class, new URIStringBinding());
        register(URL.class, new URLStringBinding());
        register(UUID.class, new UUIDStringBinding());
    }

    private void initBuiltInBindings() {

        String classResource = BasicStringBinder.class.getCanonicalName().replace('.', '/') + ".class";
        
        URL bindingClassUrl = Thread.currentThread().getContextClassLoader().getResource(classResource);
        
        String classPrefix = bindingClassUrl.toString().substring(0,
                bindingClassUrl.toString().indexOf(classResource));

        final URL builtInBindingsUrl;
        try {
            builtInBindingsUrl = new URL(classPrefix + "META-INF/bindings.xml");
        } catch (IOException e) {
            throw new IllegalStateException("Error registering bindings: " + e.getMessage(), e);
        }
        
        registerConfiguration(builtInBindingsUrl);
    }

    private void initRegisteredBindings() {
        
        Enumeration<URL> bindingsConfiguration;
        try {
            bindingsConfiguration = ClassLoaderUtils.getClassLoader().getResources("META-INF/bindings.xml");
        } catch (IOException e) {
            throw new IllegalStateException("Error registering bindings: " + e.getMessage(), e);
        }

        registerConfigurations(bindingsConfiguration);
    }

    public void registerBindingProviders(Collection<BindingProvider> providers) {

        for (BindingProvider nextProvider : providers) {
            registerBindingProvider(nextProvider);
        }
    }
    
    public void registerBindingProvider(BindingProvider provider) {
        this.bindingProviders.add(provider);
    }
    
    public void registerBindings(Collection<Binding> bindings) {

        for (Binding nextBinding : bindings) {
            registerBinding(nextBinding);
        }
    }
    
    @SuppressWarnings({"unchecked", "rawtypes"})
    public void registerBinding(Binding theBinding) {
        if (theBinding.getBindingClass().equals(theBinding.getTargetClass())) {

            StringBinding<?> binding = findBinding(theBinding.getBindingClass(), theBinding.getQualifier());
            if (binding == null) {
                throw new IllegalStateException("Targetted class " + theBinding.getBindingClass().getName() + " does not contain relevant methods");
            }
        } else if (StringBinding.class.isAssignableFrom(theBinding.getBindingClass())) {
            try {
                register(theBinding.getTargetClass(), (StringBinding)theBinding.getBindingClass().newInstance());
            } catch (InstantiationException e) {
                throw new IllegalStateException("Cannot instantiate binding class: " + theBinding.getBindingClass().getName());
            } catch (IllegalAccessException e) {
                throw new IllegalStateException("Cannot instantiate binding class: " + theBinding.getBindingClass().getName());
            }
        } else {
            registerForMethods((Class)theBinding.getTargetClass(), theBinding.getToStringMethod(), theBinding.getFromStringMethod(), theBinding.getFromStringConstructor());
        }

    }
 
    public void registerConfigurations(Enumeration<URL> bindingsConfiguration) {
        List<BindingConfiguration> configs = new ArrayList<BindingConfiguration>();
        for (URL nextLocation : IterableEnumeration.wrapEnumeration(bindingsConfiguration)) {
            
            configs.add(BindingXmlLoader.load(nextLocation));
        }
        
        for (BindingConfiguration nextConfig : configs) {
                for (Provider nextProvider : nextConfig.getProviders()) {
            
                try {
                    registerBindingProvider(nextProvider.getProviderClass().newInstance());
                } catch (InstantiationException e) {
                    throw new IllegalStateException("Cannot instantiate binding provider class: " + nextProvider.getProviderClass().getName());
                } catch (IllegalAccessException e) {
                    throw new IllegalStateException("Cannot instantiate binding provider class: " + nextProvider.getProviderClass().getName());
                }
            }
        
            registerBindings(nextConfig.getBindings());
        }
    }
    
    public void registerConfiguration(URL nextLocation) {
        
        BindingConfiguration configuration = BindingXmlLoader.load(nextLocation);

        for (Provider nextProvider : configuration.getProviders()) {
            
            try {
                registerBindingProvider(nextProvider.getProviderClass().newInstance());
            } catch (InstantiationException e) {
                throw new IllegalStateException("Cannot instantiate binding provider class: " + nextProvider.getProviderClass().getName());
            } catch (IllegalAccessException e) {
                throw new IllegalStateException("Cannot instantiate binding provider class: " + nextProvider.getProviderClass().getName());
            }
        }
        
        registerBindings(configuration.getBindings());
    }

    public <T> String convertToString(T object) {
        return convertToString(object, DefaultBinding.class);
    }

    public <T> String convertToString(T object, Class<? extends Annotation> qualifier) {
        if (object == null) {
            return null;
        }
        @SuppressWarnings("unchecked")
        Class<T> cls = (Class<T>) object.getClass();
        StringBinding<T> conv = findBinding(cls, qualifier);
        return conv.marshal(object);
    }

    public <T> T convertFromString(Class<T> cls, String str) {
        return convertFromString(cls, str, DefaultBinding.class);
    }

    public <T> T convertFromString(Class<T> cls, String str, Class<? extends Annotation> qualifier) {

        if (str == null) {
            return null;
        }
        StringBinding<T> conv = findBinding(cls, qualifier);
        return conv.unmarshal(str);
    }

    public <T> StringBinding<T> findBinding(final Class<T> cls) {
        
        return findBinding(cls, DefaultBinding.class);
    }

    public <T> StringBinding<T> findBinding(final Class<T> cls, final Class<? extends Annotation> qualifier) { 
        
        ToStringMarshaller<T> marshaller = null;
        try {
            marshaller = findMarshaller(cls, qualifier);
        } catch (IllegalStateException e) {};
        
        FromStringUnmarshaller<T> unmarshaller = null;
        try {
           unmarshaller = findUnmarshaller(cls, qualifier);
        } catch (IllegalStateException e) {};   
        
        if (marshaller != null && unmarshaller != null) {
            if (marshaller.getClass().equals(unmarshaller.getClass()) && marshaller instanceof StringBinding) {
                return (StringBinding<T>)marshaller;
            } else {
                return new CompositeStringBinding<T>(marshaller, unmarshaller);
            }
        } else if (marshaller != null) {
            return new ToStringBinding<T>(marshaller);
        } else if (unmarshaller != null) {
            return new FromStringBinding<T>(unmarshaller);
        } else { // if (marshaller == null && unmarshaller == null) {
            throw new IllegalStateException("No registered binding found: " + cls);
        } 
    }
    
    public <T> ToStringMarshaller<T> findMarshaller(final Class<T> cls) {
        return findMarshaller(cls, DefaultBinding.class);
    }
    
    @SuppressWarnings("unchecked")
    public <T> ToStringMarshaller<T> findMarshaller(final Class<T> cls, final Class<? extends Annotation> qualifier) {
        
        if (cls == null) {
            throw new IllegalArgumentException("Class must not be null");
        }
        if (qualifier.getAnnotation(BindingScope.class) == null) {
            throw new IllegalArgumentException("Qualifier must be defined using BindingScope");
        }

        Class<?> loopCls = cls;       
        ToStringMarshaller<T> conv = null;
        
        while (loopCls != Object.class) {
            
            // We check once before attempting introspection so we avoid that if possible
            ConcurrentHashMap<Class<? extends Annotation>, ToStringMarshaller<?>> qualConvs = registeredMarshalMethods.get(loopCls);
            if (qualConvs != null && qualConvs.contains(qualifier)) {
                conv = (ToStringMarshaller<T>) qualConvs.get(qualifier);
                break;
            }
            
            attachForAnnotations(cls, qualifier);
            
            qualConvs = registeredMarshalMethods.get(loopCls);
            
            if (qualConvs != null) {
                conv = (ToStringMarshaller<T>) qualConvs.get(qualifier);
                break;
            }
            
            loopCls = loopCls.getSuperclass();
        }
        if (conv == null) {
            throw new IllegalStateException("No registered marshaller found: " + cls);
        }

        return conv;
    }

    public <T> FromStringUnmarshaller<T> findUnmarshaller(final Class<T> cls) {
        return findUnmarshaller(cls, DefaultBinding.class);
    }
    
    @SuppressWarnings("unchecked")
    public <T> FromStringUnmarshaller<T> findUnmarshaller(final Class<T> cls, final Class<? extends Annotation> qualifier) {
        
        if (cls == null) {
            throw new IllegalArgumentException("Class must not be null");
        }
        if (qualifier.getAnnotation(BindingScope.class) == null) {
            throw new IllegalArgumentException("Qualifier must be defined using BindingScope");
        }

        Class<?> loopCls = cls;       
        FromStringUnmarshaller<T> conv = null;
        
        while (loopCls != Object.class) {
            
            // We check once before attempting introspection so we avoid that if possible
            ConcurrentHashMap<Class<? extends Annotation>, FromStringUnmarshaller<?>> qualConvs = registeredUnmarshalMethods.get(loopCls);
            if (qualConvs != null) {
                conv = (FromStringUnmarshaller<T>) qualConvs.get(qualifier);
                break;
            }
            
            attachForAnnotations(cls, qualifier);
            
            qualConvs = registeredUnmarshalMethods.get(loopCls);
            
            if (qualConvs != null) {
                conv = (FromStringUnmarshaller<T>) qualConvs.get(qualifier);
                break;
            }
            
            loopCls = loopCls.getSuperclass();
        }
        if (conv == null) {
            throw new IllegalStateException("No registered unmarshaller found: " + cls);
        }

        return conv;
    }
    
    private <T> void attachForAnnotations(final Class<T> cls, final Class<? extends Annotation> qualifier) {
        
        // Only apply an annotation once
        annotationsApplied.putIfAbsent(cls, new HashSet<Class<? extends Annotation>>());
        
        if (annotationsApplied.get(cls).contains(qualifier)) {
            return;
        } else {
            synchronized(this) {
                if (annotationsApplied.get(cls).contains(qualifier)) {
                    return;
                } else {
                    annotationsApplied.get(cls).add(qualifier);
                }
            }
        }
        
        Method toString = findToStringMethod(cls, qualifier);
                
        Constructor<T> con = findFromStringConstructor(cls, qualifier);
        Method fromString = findFromStringMethod(cls, qualifier);
        
        registerForMethods(cls, toString, fromString, con, qualifier);
    }

    private <T> void registerForMethods(Class<T> cls, Method toString, Method fromString, Constructor<T> con) {
        registerForMethods(cls, toString, fromString, con, DefaultBinding.class);
    }
    
    private <T> void registerForMethods(Class<T> cls, Method toString, Method fromString, Constructor<T> con, Class<? extends Annotation> qualifier) {
    
        if (toString != null) {
            
            MethodToStringMarshaller<T> toStringMarshaller = new MethodToStringMarshaller<T>(cls, toString);
            if (con != null) {
    
                ConstructorFromStringUnmarshaller<T> fromStringUnmarshaller = new ConstructorFromStringUnmarshaller<T>(con); 
                register(cls, new CompositeStringBinding<T>(toStringMarshaller, fromStringUnmarshaller), qualifier);               
            } else if (fromString != null) {
                
                MethodFromStringUnmarshaller<T> fromStringUnmarshaller = new MethodFromStringUnmarshaller<T>(cls, fromString); 
                register(cls, new CompositeStringBinding<T>(toStringMarshaller, fromStringUnmarshaller), qualifier);        
            } else {
                
                register(cls, toStringMarshaller, qualifier);
            }
        } else {
            if (con != null) {
                
                register(cls, new ConstructorFromStringUnmarshaller<T>(con), qualifier);
            } else if (fromString != null) {
                
                register(cls, new MethodFromStringUnmarshaller<T>(cls, fromString), qualifier);
            }
        }
    }
    
    private Method findToStringMethod(final Class<?> cls, final Class<? extends Annotation> qualifier) {
    
        Method result = null;
        for (BindingProvider next : bindingProviders) {
             result = next.findToStringMethod(cls, qualifier);
             if (result != null) {
                 break;
             }
        }
        return result;
    }

    private <T> Constructor<T> findFromStringConstructor(Class<T> cls, Class<? extends Annotation> qualifier) {
        
        Constructor<T> result = null;
        for (BindingProvider next : bindingProviders) {
             result = next.findFromStringConstructor(cls, qualifier);
             if (result != null) {
                 break;
             }
        }
        return result;
    }

    private Method findFromStringMethod(Class<?> cls, Class<? extends Annotation> qualifier) {

        Method result = null;
        for (BindingProvider next : bindingProviders) {
             result = next.findFromStringMethod(cls, qualifier);
             if (result != null) {
                 break;
             }
        }
        return result;
    }

    public <T> void register(final Class<T> cls, StringBinding<T> binding) {
        Class<? extends Annotation> scope = matchImplementationToScope(binding.getClass());
        register(cls, binding, scope == null ? DefaultBinding.class : scope);
    }
    
    public <T> void register(final Class<T> cls, ToStringMarshaller<T> marshaller) {
        Class<? extends Annotation> scope = matchImplementationToScope(marshaller.getClass());
        register(cls, marshaller, scope == null ? DefaultBinding.class : scope);
    }
    
    public <T> void register(final Class<T> cls, FromStringUnmarshaller<T> unmarshaller) {
        Class<? extends Annotation> scope = matchImplementationToScope(unmarshaller.getClass());
        register(cls, unmarshaller, scope == null ? DefaultBinding.class : scope);
    }
    
    public <T> void register(final Class<T> cls, StringBinding<T> binding, Class<? extends Annotation> qualifier) {
        register(cls, (ToStringMarshaller<T>) binding, qualifier);
        register(cls, (FromStringUnmarshaller<T>) binding, qualifier);
    }
    
    public <T> void register(final Class<T> cls, FromStringUnmarshaller<T> unmarshaller, Class<? extends Annotation> qualifier) {
        
        if (cls == null) {
            throw new IllegalArgumentException("Class must not be null");
        }
        if (unmarshaller == null) {
            throw new IllegalArgumentException("StringBinding must not be null");
        }

        if (qualifier == null) {
            throw new IllegalArgumentException("Qualifier must not be null");
        }

        registeredUnmarshalMethods.putIfAbsent(cls, new ConcurrentHashMap<Class<? extends Annotation>, FromStringUnmarshaller<?>>());
        ConcurrentHashMap<Class<? extends Annotation>, FromStringUnmarshaller<?>> holder = registeredUnmarshalMethods.get(cls);

        FromStringUnmarshaller<?> old = holder.put(qualifier, unmarshaller);
        if (old != null) {
            throw new IllegalStateException("Binding already registered for class: " + cls + " with qualifier: "
                    + qualifier);
        }
    }

    public <T> void register(final Class<T> cls, ToStringMarshaller<T> marshaller, Class<? extends Annotation> qualifier) {
        
        if (cls == null) {
            throw new IllegalArgumentException("Class must not be null");
        }
        if (marshaller == null) {
            throw new IllegalArgumentException("StringBinding must not be null");
        }

        if (qualifier == null) {
            throw new IllegalArgumentException("Qualifier must not be null");
        }

        registeredMarshalMethods.putIfAbsent(cls, new ConcurrentHashMap<Class<? extends Annotation>, ToStringMarshaller<?>>());
        ConcurrentHashMap<Class<? extends Annotation>, ToStringMarshaller<?>> holder = registeredMarshalMethods.get(cls);

        ToStringMarshaller<?> old = holder.put(qualifier, marshaller);
        if (old != null) {
            throw new IllegalStateException("Binding already registered for class: " + cls + " with qualifier: "
                    + qualifier);
        }
    }
    
    private <T> Class<? extends Annotation> matchImplementationToScope(Class<?> implementation) {

        for (Annotation next : implementation.getAnnotations()) {
            Class<? extends Annotation> nextType = next.annotationType();
            if (nextType.getAnnotation(BindingScope.class) != null) {
                return nextType;
            }
        }
        return null;
    }

    public Set<Class<?>> determineMarshallableClasses() {
        return Collections.unmodifiableSet(registeredMarshalMethods.keySet());
    }

    public Set<Class<?>> determineUnmarshallableClasses() {
        return Collections.unmodifiableSet(registeredUnmarshalMethods.keySet());
    }

}
