/*
 *  Copyright 2010 Chris Pheby
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.jadira.bindings.core.cdi;

import java.lang.annotation.Annotation;
import java.net.URL;
import java.util.Set;

import javax.enterprise.event.Observes;
import javax.enterprise.inject.spi.Extension;
import javax.enterprise.inject.spi.ProcessAnnotatedType;

import org.jadira.bindings.core.BasicStringBinder;
import org.jadira.bindings.core.StringBinder;
import org.jadira.bindings.core.api.FromStringUnmarshaller;
import org.jadira.bindings.core.api.StringBinding;
import org.jadira.bindings.core.api.ToStringMarshaller;
import org.jadira.bindings.core.utils.reflection.TypeHelper;

public class StringBinderExtension implements Extension, StringBinder {

    private static final StringBinder BINDING = new BasicStringBinder(); 
    
    public StringBinderExtension() {}
    
    public <T, E> void processAnnotatedType(@Observes ProcessAnnotatedType<T> pat) {

        Class<T> candidateClass = pat.getAnnotatedType().getJavaClass();

        if (!candidateClass.isInterface()) {

            try {
                if (StringBinding.class.isAssignableFrom(candidateClass)) {

                    @SuppressWarnings("unchecked")
                    Class<? extends StringBinding<E>> bindingClass = (Class<? extends StringBinding<E>>) candidateClass;
                    @SuppressWarnings("unchecked")
                    Class<E> boundType = (Class<E>) TypeHelper.getTypeArguments(StringBinding.class, bindingClass).get(
                            0);
                    BINDING.register(boundType, bindingClass.newInstance());
                } else if (ToStringMarshaller.class.isAssignableFrom(candidateClass)
                        || FromStringUnmarshaller.class.isAssignableFrom(candidateClass)) {
                    if (ToStringMarshaller.class.isAssignableFrom(candidateClass)) {

                        @SuppressWarnings("unchecked")
                        Class<? extends ToStringMarshaller<E>> bindingClass = (Class<? extends ToStringMarshaller<E>>) candidateClass;
                        @SuppressWarnings("unchecked")
                        Class<E> boundType = (Class<E>) TypeHelper.getTypeArguments(ToStringMarshaller.class,
                                bindingClass).get(0);
                        BINDING.register(boundType, bindingClass.newInstance());
                    }
                    if (FromStringUnmarshaller.class.isAssignableFrom(candidateClass)) {

                        @SuppressWarnings("unchecked")
                        Class<? extends FromStringUnmarshaller<E>> bindingClass = (Class<? extends FromStringUnmarshaller<E>>) candidateClass;
                        @SuppressWarnings("unchecked")
                        Class<E> boundType = (Class<E>) TypeHelper.getTypeArguments(FromStringUnmarshaller.class,
                                bindingClass).get(0);
                        BINDING.register(boundType, bindingClass.newInstance());
                    }
                } else {
                    try {
                        BINDING.findBinding(candidateClass);
                    } catch (IllegalStateException e) {
                        // Do nothing - this just meant we can't bind this class
                    }
                }
            } catch (InstantiationException e) {
                throw new IllegalStateException("Cannot instantiate binding class: " + candidateClass);
            } catch (IllegalAccessException e) {
                throw new IllegalStateException("Cannot instantiate binding class: " + candidateClass);
            }
        }
    }

    public void registerConfiguration(URL nextLocation) {
        BINDING.registerConfiguration(nextLocation);
    }

    public <T> String convertToString(T object) {
        return BINDING.convertToString(object);
    }

    public <T> String convertToString(T object, Class<? extends Annotation> qualifier) {
        return BINDING.convertToString(object, qualifier);
    }

    public <T> T convertFromString(Class<T> cls, String str) {
        return BINDING.convertFromString(cls, str);
    }

    public <T> T convertFromString(Class<T> cls, String str, Class<? extends Annotation> qualifier) {
        return BINDING.convertFromString(cls, str, qualifier);
    }

    public <T> StringBinding<T> findBinding(Class<T> cls) {
        return BINDING.findBinding(cls);
    }

    public <T> ToStringMarshaller<T> findMarshaller(Class<T> cls) {
        return BINDING.findMarshaller(cls);
    }

    public <T> FromStringUnmarshaller<T> findUnmarshaller(Class<T> cls) {
        return BINDING.findUnmarshaller(cls);
    }

    public <T> StringBinding<T> findBinding(Class<T> cls, Class<? extends Annotation> qualifier) {
        return BINDING.findBinding(cls, qualifier);
    }

    public <T> ToStringMarshaller<T> findMarshaller(Class<T> cls, Class<? extends Annotation> qualifier) {
        return BINDING.findMarshaller(cls, qualifier);
    }

    public <T> FromStringUnmarshaller<T> findUnmarshaller(Class<T> cls, Class<? extends Annotation> qualifier) {
        return BINDING.findUnmarshaller(cls, qualifier);
    }

    public <T> void register(Class<T> cls, StringBinding<T> converter) {
        BINDING.register(cls, converter);
    }

    public <T> void register(Class<T> cls, ToStringMarshaller<T> converter) {
        BINDING.register(cls, converter);
    }

    public <T> void register(Class<T> cls, FromStringUnmarshaller<T> converter) {
        BINDING.register(cls, converter);
    }

    public <T> void register(Class<T> cls, StringBinding<T> converter, Class<? extends Annotation> qualifier) {
        BINDING.register(cls, converter, qualifier);
    }

    public <T> void register(Class<T> cls, ToStringMarshaller<T> converter, Class<? extends Annotation> qualifier) {
        BINDING.register(cls, converter, qualifier);
    }

    public <T> void register(Class<T> cls, FromStringUnmarshaller<T> converter, Class<? extends Annotation> qualifier) {
        BINDING.register(cls, converter, qualifier);
    }

    public Set<Class<?>> determineMarshallableClasses() {
        return BINDING.determineMarshallableClasses();
    }

    public Set<Class<?>> determineUnmarshallableClasses() {
        return BINDING.determineUnmarshallableClasses();
    }
}