/*
 *  Copyright 2010 Chris Pheby
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.jadira.bindings.core.general;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Modifier;

import org.jadira.bindings.core.api.FromStringUnmarshaller;

/**
 * Unmarshaller supports constructor parameterised by a String
 * @param <T> the type of the class for the conversion
 */
public final class ConstructorFromStringUnmarshaller<T> implements FromStringUnmarshaller<T> {

    private final Constructor<T> unmarshal;

    private final Class<T> targetClass;
    
    public ConstructorFromStringUnmarshaller(Constructor<T> unmarshal) {
        
        this.targetClass = unmarshal.getDeclaringClass();
        
        if (getTargetClass().isInterface() 
                || Modifier.isAbstract(getTargetClass().getModifiers()) 
                || getTargetClass().isLocalClass() 
                || getTargetClass().isMemberClass()) {
            throw new IllegalStateException("unmarshal constructor must have an instantiable target class");
        }
        
        this.unmarshal = unmarshal;
    }

    /**
     * Unmarshal String to an instance of the bound class
     */
    public T unmarshal(String str) {
        try {
            return unmarshal.newInstance(str);
        } catch (IllegalAccessException ex) {
            throw new IllegalStateException("Constructor is not accessible");
        } catch (InstantiationException ex) {
            throw new IllegalStateException("Constructor is not valid");
        } catch (InvocationTargetException ex) {
            if (ex.getCause() instanceof RuntimeException) {
                throw (RuntimeException) ex.getCause();
            }
            throw new RuntimeException(ex.getMessage(), ex.getCause());
        }
    }
    
    /**
     * Obtain the bound class
     * @return Class<T>
     */
    protected Class<T> getTargetClass() {
        return targetClass;
    }
}
