/*
 *  Copyright 2010 Chris Pheby
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distringibuted under the License is distringibuted on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.jadira.bindings.core.general;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import org.jadira.bindings.core.api.FromStringUnmarshaller;

/**
 * String binding that supports an unmarshal method. The
 * unmarshal method must be statically scoped. It must accept a single parameter
 * of type T and return a type of String. For example: </p>
 * <p>
 * {@code public static BoundType unmarshal(String string)}
 * </p>
 * @param <T> the type of the class for the conversion
 */
public final class MethodFromStringUnmarshaller<T> implements FromStringUnmarshaller<T> {

    private final Method unmarshal;

    private final Class<T> targetClass;
    
    public MethodFromStringUnmarshaller(Class<T> targetClass, Method unmarshal) {

        this.targetClass = targetClass;
        
        if (unmarshal.getParameterTypes().length != 1) {
            throw new IllegalStateException("unmarshal method must define a single parameter");
        }
        if (!Modifier.isStatic(unmarshal.getModifiers())) {
            throw new IllegalStateException("unmarshal method must be defined as static");
        }

        if (unmarshal.getParameterTypes()[0] != String.class) {
            throw new IllegalStateException("unmarshal method must be parameterized by String");
        }
        if (!targetClass.isAssignableFrom(unmarshal.getReturnType())) {
            throw new IllegalStateException("unmarshal method must return " + targetClass.getSimpleName());
        }

        this.unmarshal = unmarshal;
    }

    /**
     * Unmarshal the given String to an instance of the bound class
     */
    public T unmarshal(String string) {

        try {
            return getTargetClass().cast(unmarshal.invoke(null, string));
        } catch (IllegalAccessException ex) {
            throw new IllegalStateException("Method is not accessible");
        } catch (InvocationTargetException ex) {
            if (ex.getCause() instanceof RuntimeException) {
                throw (RuntimeException) ex.getCause();
            }
            throw new RuntimeException(ex.getMessage(), ex.getCause());
        }
    }
    
    /**
     * Obtain the bound class
     * @return Class<T>
     */
    protected Class<T> getTargetClass() {
        return targetClass;
    }
}
