/*
 *  Copyright 2010 Chris Pheby
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.jadira.bindings.core.general;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import org.jadira.bindings.core.api.ToStringMarshaller;

/**
 * Base class providing capability to perform marshalling of target object type
 * to String. This class uses reflection.
 * <p>
 * The marshal method must either
 * </p>
 * <p>
 * a) be instance scoped and defined as part of class T. It must accept no
 * parameters and return a type of String. For example:
 * </p>
 * <p>
 * {@code public String marshal()}
 * </p>
 * <p>
 * b) be statically scoped. It must accept a single parameter of type T and
 * return a type of String. For example:
 * </p>
 * <p>
 * {@code public static String marshal(BoundType param)}
 * </p>
 * @param <T> the type of the converter
 */
public class MethodToStringMarshaller<T> implements ToStringMarshaller<T> {

    private final Class<T> targetClass;

    private final Method marshal;

    /**
     * Create a new instance
     * @param targetClass Target class
     * @param marshal Marshal instance method on the target class
     */
    public MethodToStringMarshaller(Class<T> targetClass, Method marshal) {
        
        if (marshal.getParameterTypes().length == 0 && Modifier.isStatic(marshal.getModifiers())) {
            throw new IllegalStateException("marshal method must either be instance scope or define a single parameter");
        } else if (marshal.getParameterTypes().length == 1 && (!Modifier.isStatic(marshal.getModifiers()))) {
            throw new IllegalStateException("marshal method must either be instance scope or define a single parameter");
        } else if (marshal.getParameterTypes().length >= 2) {
            throw new IllegalStateException("marshal method must either be instance scope or define a single parameter");
        }
        
        if (marshal.getReturnType() != String.class) {
            throw new IllegalStateException("marshal method must return a String");
        }
        if (!marshal.getDeclaringClass().isAssignableFrom(targetClass) && !Modifier.isStatic(marshal.getModifiers())) {
            throw new IllegalStateException("marshal method must be defined as part of " + targetClass.getSimpleName());
        }

        this.targetClass = targetClass;
        this.marshal = marshal;
    }

    /**
     * Marshal the target object to a String using the marshal method
     */
    public String marshal(T object) {

        try {
            if (Modifier.isStatic(marshal.getModifiers())) {
                return (String) marshal.invoke(null, object);
            } else {
                return (String) marshal.invoke(object);
            }
        } catch (IllegalAccessException ex) {
            throw new IllegalStateException("Method is not accessible");
        } catch (InvocationTargetException ex) {
            if (ex.getCause() instanceof RuntimeException) {
                throw (RuntimeException) ex.getCause();
            }
            throw new RuntimeException(ex.getMessage(), ex.getCause());
        }
    }
   
    /**
     * Obtain the bound class
     * @return Class<T>
     */
    protected Class<T> getTargetClass() {
        return targetClass;
    }
}
