/*
 *  Copyright 2010 Christopher Pheby
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.jadira.bindings.core.loader;

import java.io.IOException;
import java.io.InputStream;
import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.jadira.bindings.core.annotation.BindingScope;
import org.jadira.bindings.core.annotation.DefaultBinding;
import org.jadira.bindings.core.spi.BindingProvider;
import org.jadira.bindings.core.utils.reflection.ClassLoaderUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public final class BindingXmlLoader {

    private BindingXmlLoader() {
    }

    public static BindingConfiguration load(URL location) {
        Document doc;
        try {
            doc = loadDocument(location);
        } catch (IOException e) {
            throw new IllegalStateException("Cannot load " + location.toExternalForm(), e);
        } catch (ParserConfigurationException e) {
            throw new IllegalStateException("Cannot initialise parser for " + location.toExternalForm(), e);
        } catch (SAXException e) {
            throw new IllegalStateException("Cannot parse " + location.toExternalForm(), e);
        }
        BindingConfiguration configuration = parseDocument(location, doc);
        return configuration;
    }

    private static Document loadDocument(URL location) throws IOException, ParserConfigurationException, SAXException {

        InputStream inputStream = null;

        if (location != null) {
            URLConnection urlConnection = location.openConnection();
            urlConnection.setUseCaches(false);
            inputStream = urlConnection.getInputStream();
        }
        if (inputStream == null) {
            throw new IOException("Failed to obtain InputStream for named location: " + location.toExternalForm());
        }

        InputSource inputSource = new InputSource(inputStream);

        List<SAXParseException> errors = new ArrayList<SAXParseException>();
        DocumentBuilder docBuilder = constructDocumentBuilder(errors);

        Document document = docBuilder.parse(inputSource);
        if (!errors.isEmpty()) {
            throw new IllegalStateException("Invalid file: " + location.toExternalForm(), (Throwable) errors.get(0));
        }
        return document;
    }

    private static DocumentBuilder constructDocumentBuilder(List<SAXParseException> errors)
            throws ParserConfigurationException {

        DocumentBuilderFactory documentBuilderFactory = constructDocumentBuilderFactory();
        DocumentBuilder docBuilder = documentBuilderFactory.newDocumentBuilder();
        docBuilder.setEntityResolver(new BindingXmlEntityResolver());
        docBuilder.setErrorHandler(new BindingXmlErrorHandler(errors));
        return docBuilder;
    }

    private static DocumentBuilderFactory constructDocumentBuilderFactory() {

        DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();

        documentBuilderFactory.setValidating(true);
        documentBuilderFactory.setNamespaceAware(true);

        try {
            documentBuilderFactory.setAttribute("http://apache.org/xml/features/validation/schema", true);
        } catch (IllegalArgumentException e) {
            // Ignore
        }
        documentBuilderFactory.setAttribute("http://java.sun.com/xml/jaxp/properties/schemaLanguage",
                "http://www.w3.org/2001/XMLSchema");
        documentBuilderFactory.setAttribute("http://java.sun.com/xml/jaxp/properties/schemaSource",
                "classpath:/jadira-bindings.xsd");
        return documentBuilderFactory;
    }

    private static BindingConfiguration parseDocument(URL location, Document doc) {

        BindingConfiguration result = new BindingConfiguration();

        Element docRoot = doc.getDocumentElement();
        for (Node next : IterableNodeList.make(docRoot.getChildNodes())) {
            if (Node.ELEMENT_NODE == next.getNodeType()) {
                Element element = (Element) next;

                if ("http://org.jadira.bindings/xml/ns/stringbinding".equals(element.getNamespaceURI())
                        && "provider".equals(element.getLocalName())) {

                    Provider provider = parseProviderElement(location, element);
                    result.addProvider(provider);
                }

                if ("http://org.jadira.bindings/xml/ns/stringbinding".equals(element.getNamespaceURI())
                        && "binding".equals(element.getLocalName())) {

                    Binding binding = parseBindingElement(location, element);
                    result.addBinding(binding);
                }
            }
        }
        return result;
    }

    private static Provider parseProviderElement(URL location, Element element) {

        Class<?> providerClass = lookupClass(element.getAttribute("class"));

        if (providerClass == null) {
            throw new IllegalStateException("Referenced class {" + element.getAttribute("class")
                    + "} could not be found");
        }
        if (!BindingProvider.class.isAssignableFrom(providerClass)) {
            throw new IllegalStateException("Referenced class {" + element.getAttribute("class")
                    + "} did not implement BindingProvider");
        }

        @SuppressWarnings("unchecked")
        final Class<BindingProvider> typedProviderClass = (Class<BindingProvider>) providerClass;
        return new Provider((Class<BindingProvider>) typedProviderClass);
    }

    private static Binding parseBindingElement(URL location, Element element) {

        Method toStringMethod = null;
        Method fromStringMethod = null;
        Constructor<?> fromStringConstructor = null;

        Class<?> bindingClass = lookupClass(element.getAttribute("class"));
        if (bindingClass == null) {
            throw new IllegalStateException("Referenced class {" + element.getAttribute("class")
                    + "} could not be found");
        }

        Class<?> targetClass = lookupClass(element.getAttribute("targetClass"));
        if (targetClass == null) {
            targetClass = bindingClass;
        }

        Class<? extends Annotation> qualifier = DefaultBinding.class;
        if (element.getAttribute("qualifier").length() > 0) {
            try {
                @SuppressWarnings("unchecked")
                final Class<? extends Annotation> myQualifier = (Class<? extends Annotation>) Class.forName(element
                        .getAttribute("qualifier"));
                qualifier = myQualifier;

                if (qualifier.getAnnotation(BindingScope.class) == null) {
                    throw new IllegalStateException("Qualifier class {" + element.getAttribute("qualifier")
                            + "} was not marked as BindingScope");
                }
            } catch (ClassNotFoundException e) {
                throw new IllegalStateException("Qualifier class {" + element.getAttribute("qualifier")
                        + "} could not be found");
            }
        }

        for (Node next : IterableNodeList.make(element.getChildNodes())) {
            if (Node.ELEMENT_NODE == next.getNodeType()) {
                Element childElement = (Element) next;
                if ("http://org.jadira.bindings/xml/ns/stringbinding".equals(element.getNamespaceURI())
                        && "toStringMethod".equals(element.getLocalName())) {

                    String toStringMethodName = childElement.getTextContent();

                    try {
                        toStringMethod = bindingClass.getMethod(toStringMethodName, new Class[] { targetClass });
                    } catch (SecurityException e) {
                    } catch (NoSuchMethodException e) {
                    }
                    if (!String.class.equals(toStringMethod.getReturnType())
                            || !Modifier.isStatic(toStringMethod.getModifiers())) {
                        toStringMethod = null;
                    }
                    if (toStringMethod == null && bindingClass.equals(targetClass)) {
                        try {
                            toStringMethod = bindingClass.getMethod(toStringMethodName, new Class[] {});
                        } catch (SecurityException e) {
                        } catch (NoSuchMethodException e) {
                        }
                        if (Modifier.isStatic(toStringMethod.getModifiers())) {
                            toStringMethod = null;
                        }
                    }

                } else if ("http://org.jadira.bindings/xml/ns/stringbinding".equals(element.getNamespaceURI())
                        && "fromStringMethod".equals(element.getLocalName())) {

                    String fromStringMethodName = childElement.getTextContent();

                    try {
                        fromStringMethod = bindingClass.getMethod(fromStringMethodName, new Class[] { String.class });
                    } catch (SecurityException e) {
                    } catch (NoSuchMethodException e) {
                    }
                    if (!targetClass.isAssignableFrom(fromStringMethod.getReturnType())
                            || !Modifier.isStatic(fromStringMethod.getModifiers())) {
                        fromStringMethod = null;
                    }

                } else if ("http://org.jadira.bindings/xml/ns/stringbinding".equals(element.getNamespaceURI())
                        && "fromStringConstructor".equals(element.getLocalName())) {

                    try {
                        fromStringConstructor = bindingClass.getConstructor(new Class[] { String.class });
                    } catch (SecurityException e) {
                    } catch (NoSuchMethodException e) {
                    }
                }
            }
        }

        return new Binding(bindingClass, targetClass, qualifier, toStringMethod, fromStringMethod,
                fromStringConstructor);
    }

    private static final Class<?> lookupClass(String elementName) {

        Class<?> clazz = null;
        try {
            clazz = ClassLoaderUtils.getClassLoader().loadClass(elementName);
        } catch (ClassNotFoundException e) {
            return null;
        }

        return clazz;
    }

    private static class BindingXmlErrorHandler implements ErrorHandler {

        private List<SAXParseException> errors;

        BindingXmlErrorHandler(List<SAXParseException> errors) {
            this.errors = errors;
        }

        public void error(SAXParseException error) {
            errors.add(error);
        }

        public void fatalError(SAXParseException error) {
            errors.add(error);
        }

        public void warning(SAXParseException warn) {
            // ignore
        }
    }

    /**
     * Helper used for working with the JDKs classloader API
     */
    private static class IterableNodeList implements Iterable<Node> {

        private final NodeList nodeList;

        public IterableNodeList(NodeList nodeList) {
            this.nodeList = nodeList;
        }

        public Iterator<Node> iterator() {

            return new Iterator<Node>() {

                private int currentIndex = 0;

                public boolean hasNext() {
                    return currentIndex < nodeList.getLength();
                }

                public Node next() {
                    Node next = nodeList.item(currentIndex);
                    currentIndex++;
                    if (next == null) {
                        throw new NoSuchElementException("No element found for index: " + currentIndex);
                    }
                    return next;
                }

                public void remove() {
                    throw new UnsupportedOperationException("IterableNodeList does not support remove");
                }
            };
        }

        private static Iterable<Node> make(NodeList enumeration) {
            return new IterableNodeList(enumeration);
        }
    }
}
