/*
 *  Copyright 2010 Chris Pheby
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.jadira.bindings.core.provider;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;

import org.jadira.bindings.core.annotation.BindingScope;
import org.jadira.bindings.core.annotation.DefaultBinding;
import org.jadira.bindings.core.utils.reflection.TypeHelper;

/**
 * Used to implement a binding provider instance that uses annotations to match methods 
 * @param <T> The annotation used to identify a toString method
 * @param <F> The annotation used to identify a fromString method
 */
public class AbstractAnnotationMatchingBindingProvider<T extends Annotation, F extends Annotation> {

    public Method findToStringMethod(final Class<?> cls, final Class<? extends Annotation> qualifier) {

        @SuppressWarnings("unchecked")
        Class<T> toStringAnnotation = (Class<T>) TypeHelper.getTypeArguments(
                AbstractAnnotationMatchingBindingProvider.class, this.getClass()).get(0);

        Method matched = null;
        Class<?> loopCls = cls;

        while (loopCls != Object.class && matched == null) {
            Method[] methods = loopCls.getDeclaredMethods();
            for (Method method : methods) {
                T toString = method.getAnnotation(toStringAnnotation);
                if (toString != null) {
                    if (matchMethodToScope(qualifier, method.getAnnotations())) {
                        if (matched != null) {
                            throw new IllegalStateException("Two methods are annotated with @ToString for scope "
                                    + qualifier.getSimpleName());
                        }
                        matched = method;
                    }
                }
            }
            loopCls = loopCls.getSuperclass();
        }
        return matched;
    }

    public <C> Constructor<C> findFromStringConstructor(Class<C> cls, Class<? extends Annotation> qualifier) {

        @SuppressWarnings("unchecked")
        Class<F> fromStringAnnotation = (Class<F>) TypeHelper.getTypeArguments(
                AbstractAnnotationMatchingBindingProvider.class, this.getClass()).get(1);

        try {
            Constructor<C> con = cls.getDeclaredConstructor(String.class);
            F fromString = con.getAnnotation(fromStringAnnotation);

            boolean matchedScope = false;
            if (fromString != null) {

                matchedScope = matchMethodToScope(qualifier, con.getAnnotations());
            }
            if (fromString != null && matchedScope) {
                return con;
            }
            return null;
        } catch (NoSuchMethodException ex) {
            return null;
        }
    }

    public Method findFromStringMethod(Class<?> cls, Class<? extends Annotation> qualifier) {

        @SuppressWarnings("unchecked")
        Class<F> fromStringAnnotation = (Class<F>) TypeHelper.getTypeArguments(
                AbstractAnnotationMatchingBindingProvider.class, this.getClass()).get(1);

        Method matched = null;
        Class<?> loopCls = cls;

        while (loopCls != Object.class && matched == null) {
            Method[] methods = loopCls.getDeclaredMethods();
            for (Method method : methods) {
                if (cls.isAssignableFrom(method.getReturnType())) {
                    F fromString = method.getAnnotation(fromStringAnnotation);
                    if (fromString != null) {
                        if (matchMethodToScope(qualifier, method.getAnnotations())) {
                            if (matched != null) {
                                throw new IllegalStateException("Two methods are annotated with @FromString for scope "
                                        + qualifier.getSimpleName());
                            }
                            matched = method;
                        }
                    }
                }

            }
            loopCls = loopCls.getSuperclass();
        }
        return matched;
    }

    /**
     * Returns true if a method can be matched for the required scope
     * @param qualifier The qualifier that the method must belong to
     * @param annotations Array of Annotations on the target method
     * @return True if method can be matched for the given scope
     */
    private boolean matchMethodToScope(Class<? extends Annotation> qualifier, Annotation[] annotations) {

        boolean scopeDeclared = false;
        boolean matchedScope = false;

        for (Annotation next : annotations) {
            Class<? extends Annotation> nextType = next.annotationType();
            if (nextType.getAnnotation(BindingScope.class) != null) {
                scopeDeclared = true;
                if (nextType.equals(qualifier)) {
                    matchedScope = true;
                    break;
                }
            }
        }

        if (!scopeDeclared && DefaultBinding.class.equals(qualifier)) {
            matchedScope = true;
        }

        return matchedScope;
    }

}
