/*
 *  Copyright 2010 Chris Pheby
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.jadira.bindings.core.spring;

import java.lang.annotation.Annotation;
import java.util.HashSet;
import java.util.Set;

import org.jadira.bindings.core.BasicStringBinder;
import org.jadira.bindings.core.StringBinder;
import org.jadira.bindings.core.annotation.BindingScope;
import org.jadira.bindings.core.annotation.DefaultBinding;
import org.springframework.core.convert.TypeDescriptor;
import org.springframework.core.convert.converter.ConditionalGenericConverter;

public class StringBindingConverter implements ConditionalGenericConverter {

    private static final StringBinder BINDING = new BasicStringBinder();
    
    public Set<ConvertiblePair> getConvertibleTypes() {
        
        Set<ConvertiblePair> result = new HashSet<ConvertiblePair>();
        
        Set<Class<?>> marshallableClasses = BINDING.determineMarshallableClasses();
        for (Class<?> next : marshallableClasses) {
            result.add(new ConvertiblePair(next, String.class));
        }
        
        Set<Class<?>> unmarshallableClasses = BINDING.determineUnmarshallableClasses();
        for (Class<?> next : unmarshallableClasses) {
            result.add(new ConvertiblePair(String.class, next));
        }
        
        return result;
    }

    public Object convert(Object source, TypeDescriptor sourceType, TypeDescriptor targetType) {
        
        final Object result;
        if (source instanceof String) {
            result = BINDING.convertFromString(targetType.getObjectType(), (String)source, matchAnnotationToScope(targetType.getAnnotations()));
        } else {
            result = BINDING.convertToString(source, matchAnnotationToScope(sourceType.getAnnotations()));
        }
        return result;
    }

    public boolean matches(TypeDescriptor sourceType, TypeDescriptor targetType) {

        if (String.class.equals(sourceType.getObjectType())) {
            try {
                BINDING.findUnmarshaller(targetType.getObjectType(), matchAnnotationToScope(targetType.getAnnotations()));
                return true;
            } catch (IllegalStateException e) {
                return false;
            }
        } else {
            try {
                BINDING.findMarshaller(sourceType.getObjectType(), matchAnnotationToScope(sourceType.getAnnotations()));
                return true;
            } catch (IllegalStateException e) {
                return false;
            }
        }
    }
    
    private <T> Class<? extends Annotation> matchAnnotationToScope(Annotation[] annotations) {

        for (Annotation next : annotations) {
            Class<? extends Annotation> nextType = next.annotationType();
            if (nextType.getAnnotation(BindingScope.class) != null) {
                return nextType;
            }
        }
        return DefaultBinding.class;
    }
}
